
-- UnityEngine
Time = UnityEngine.Time
Vector2 = UnityEngine.Vector2
Vector3 = UnityEngine.Vector3
Transform = UnityEngine.Transform
Quaternion = UnityEngine.Quaternion
GameObject = UnityEngine.GameObject
Object = UnityEngine.Object
Random = UnityEngine.Random
Application = UnityEngine.Application
Input = UnityEngine.Input
Screen = UnityEngine.Screen
KeyCode = UnityEngine.KeyCode
WaitForSeconds = UnityEngine.WaitForSeconds
SceneManager = UnityEngine.SceneManagement.SceneManager

-- Fenix <csharp>
MathUtils = Fenix.MathUtils
WaitForInstantiate = Fenix.Assets.Helper.WaitForInstantiate
ObjectCaches = Fenix.ObjectCaches

-- Fenix <lua>
class = require("base.class")
enum = require("base.enum")
locale = require("base.locale")
L = locale
logger = require("base.logger")
functor = require("base.functor")
startcoroutine = function (fn) Fenix.App.GetInstance():StartCoroutine(fn) end
CreateFrame = require("ui.frame").CreateFrame

if not table.copy_from_once then 
    table.copy_from_once = function (t, f, k)
        if type(k) == "table" then 
            for _, v in ipairs(k) do 
                if not t[v] then 
                    t[v] = f[v]
                end
            end
        else
            if not t[k] then 
                t[k] = f[k]
            end
        end
    end
end

-- lua extensions
if not string.split then 
    string.split = function (strVal, delim)
        assert(type(delim) == "string" and string.len(delim) > 0, "bad delimiter")
        local start = 1
        local results = {}
        -- find each instance of a string followed by the delimiter
        while true do
            local pos = string.find(strVal, delim, start, true) -- plain find
            if not pos then
                break
            end
            table.insert(results, string.sub(strVal, start, pos - 1))
            start = pos + string.len(delim)
        end 
        -- insert final one(after last delimiter)
        table.insert(results, string.sub(strVal, start))
        return results
    end 
end

if not string.toVec2 then 
    string.toVec2 = function (strVal)
        local xy = string.split(strVal, ",")
        xy[1] = tonumber(xy[1])
        xy[2] = tonumber(xy[2])
        return xy
    end
end

if not string.toVec3 then 
    string.toVec3 = function (strVal)
        local xy = string.split(strVal, ",")
        xy[1] = tonumber(xy[1])
        xy[2] = tonumber(xy[2])
        xy[3] = tonumber(xy[3])
        return xy
    end
end
-- print(unpack(string.toVec3("1,2,3")))

if not string.cache then 
    local cache = {}
    string.cache = function (strVal)
        if strVal == nil then 
            return -1 
        end 
        local index = cache[strVal]
        if index == nil then
            index = ObjectCaches.AddSharedObject(strVal)
            cache[strVal] = index
        end
        return index
    end
end

-- 用法范例:
-- local data = { name = "Tom", age = 20, city = "Shanghai" }
-- print(string.replace_var("你好, 我是{name}. 今年{age}岁. 住在{city}.", function (name)
--     return data[name]
-- end))
-- print(string.replace_var("test ((name)) = (name).", function (name)
--     return "John"
-- end, {"(", ")"}))
if not string.replace_var then 
    string.replace_var = function (text, handler, enclosures)
        return (string.gsub(text, enclosures or "%b()", function (name)
            return handler(string.sub(name, 2, -2))
        end))
    end
end

if not string.replace_varobj then 
    string.replace_varobj = function (text, obj, enclosures)
        return string.replace_var(text, function (name) 
            return obj[name]
        end, enclosures)
    end
end

if not math.epsilon then 
    math.epsilon = 0.001
end

if not math.rotate then 
    math.rotate = function (vec, degree)
        local angle = math.rad(-degree)
        return { 
            vec[1] * math.cos(angle) - vec[2] * math.sin(angle), 
            vec[1] * math.sin(angle) + vec[2] * math.cos(angle)
        }
    end
end

if not math.div then 
    math.div = function (a, b)
        return b == 0 and 1 or (a / b)
    end
end

if not math.clamp then 
    math.clamp = function (value, min, max)
        return value < min and min or (value > max and max or value)
    end
end

if not debug.log then 
    debug.log = logger.get("Game", "Debug")
end

if not debug.profile then 
    debug.profile = function (title, call)
        local time = Time.realtimeSinceStartup
        call()
        print(title, "in", (Time.realtimeSinceStartup - time) * 1000, "ms")
    end
end

table.copy_from_once(debug, require("base.test"), {"testable", "runtests", "deprecated"})

local __uid = 0
local __layers = {}

local __getLayer = function (name)
    local find = __layers[name]
    if find == nil then 
        find = UnityEngine.LayerMask.NameToLayer(name)
        __layers[name] = find
    end
    return find
end

return {
    id = function ()
        __uid = __uid + 1
        return __uid
    end, 

    decVec3 = function (x, y, z)
        return Vector3(x * 0.0001, y * 0.0001, z * 0.0001)
    end, 

    i2f = function (n)
        return n * 0.0001
    end, 

    f2i = function (n)
        return math.ceil(n * 10000)
    end, 

    capsule = function ()
        return GameObject.CreatePrimitive(UnityEngine.PrimitiveType.Capsule)
    end, 

    cube = function ()
        return GameObject.CreatePrimitive(UnityEngine.PrimitiveType.Cube)
    end, 

    destroy = function (...)
        for i = 1, select("#", ...) do 
            local obj = select(i, ...)
            if obj and obj.destroy then 
                pcall(function ()
                    obj:destroy()
                end)
            end
        end
    end, 

    strings = function (tb, default_val)
        return setmetatable(tb, { 
            __index = function (self, index)
                return default_val
            end, 
        })
    end, 

    -- 返回一个安全函数表, 对该表取无效值将不会产生执行失败
    functions = function (tb, err)
        return setmetatable(tb, { 
            __index = function (self, index)
                return function () 
                    if err then 
                        logger.error("you've called an invalid script:", index)
                    end
                end
            end, 
        })
    end, 

    mask = function (...)
        local lv = 0
        local sbor = Fenix.ConvertUtils.sbor
        for i = 1, select("#", ...) do 
            lv = sbor(lv, __getLayer((select(i, ...))))
        end
        return lv
    end, 

    layers = setmetatable({}, {
        __index = function (self, name)
            return __getLayer(name)
        end, 

        __call = function (self, ...)
            local lv = 0
            local sbor = Fenix.ConvertUtils.sbor
            for i = 1, select("#", ...) do 
                lv = sbor(lv, __getLayer((select(i, ...))))
            end
            return lv
        end, 
    }), 

    -- 重新载入
    reload = function (name)
        package.loaded[name] = nil
        package.preloaded[name] = nil
        return require(name)
    end, 

    -- 载入模块（即使不存在也不会报错，直接返回nil)
    require = function (name)
        local status, results = pcall(function (mod) return require(mod) end, name)
        -- print(name, status, results)
        return status and results or nil
    end, 
}
