local ButtonWrapper = require("ui.button")
local ImageWrapper = require("ui.image")
local DropdownWrapper = require("ui.dropdown")
local InputFieldWrapper = require("ui.inputfield")
local WidgetWrapper = require("ui.widget")

return class {
    extensions = "ui/frame+handler.lua", 
    --__unsafe = true, 

    -- params:
    --     parent 被创建frame的父节点(gameObject/transform), 同时也是 Frame 组件的容器
    ctor = function (self, parent, name)
        if parent then 
            self.__loader = parent:AddComponent(Fenix.Assets.PrefabLoader)
        end
        self.name = name
        -- 通过 Find 查询的所有元素的 cache 记录 (结果经过 lua 封装, 不一定是直接本地类实例)
        self._transforms = {}

        -- 记录界面上注册的事件，界面销毁时反注册
        self.rec_events = {
            click = {},
            drag = {},
            drop = {},
            updown = {},
            single = {},
            double = {},
            hover = {},
            singleclick = {},
            scroll = {},
        }

        self:onInit()
    end, 

    visible = {
        getter = function (self)
            return self.__loader.enabled
        end, 
        setter = function (self, value)
            self.__loader.enabled = value
        end, 
    }, 

    -- static
    _new = function (cls, args)
        local inst = nil
        if type(args.parent) == "string" then 
            local parent = UnityEngine.GameObject.Find(args.parent)
            inst = cls.new(parent, args.name)
        else
            inst = cls.new(args.parent, args.name)
        end
        
        if inst then
            inst.gameObject = args.gameObject
            if args.events then 
                for k, v in pairs(args.events) do 
                    inst:on(k, v)
                end
            end
            inst.args = args
            inst.owner = args.owner            
            if args.asset then
                -- inst:load(args.asset)
                -- edit by 张浩 start
                app:startCoroutine(function()
                    coroutine.yield(nil)
                    inst:load(args.asset)
                end) 
                -- edit by 张浩 end
            end
        end
        return inst
    end, 

    AddComponent = function (self, type)
        return self.gameObject:AddComponent(type)
    end, 

    -- 静态方法, 创建指定的 frame 类
    -- template: string
    -- parent: gameObject/transform/string
    -- name: string
    -- asset: string
    CreateFrame = function (args)
        local cls = require(args.class or args.template)
        return cls:_new(args)
    end,

    load = function (self, assetPath)
        if self.__loader then 
            self.__loader:Load(assetPath, function (gameObject)
                --[[
                self._error = gameObject == nil
                self.gameObject = gameObject
                self.transform = gameObject and gameObject.transform
                self:onLoaded()
                self.is_loaded = true
                ]]
                -- edit by 张浩   start
                if self:willBeDisplay() then
                    self._error = gameObject == nil
                    self.gameObject = gameObject
                    self.transform = gameObject and gameObject.transform
                    self:onLoaded()
                    self.is_loaded = true
                    self:excuteCommand()
                else
                    self:destroy()
                end
                -- edit by 张浩   end
                
            end)
        end
    end, 

    -- 加载完毕后，判断当前界面是否继续显示
    willBeDisplay = function(self)
        local vsb, frame = frame_manager:isDisplay(self.name)
        if not vsb or (frame and frame ~= self) then
            return false
        end
        return true
    end,

    destroy = function (self)
        if self.__loader then 
            app:startCoroutine(
                function()
                    coroutine.yield(nil)
                    if self.__loader then
                        UnityEngine.Object.DestroyImmediate(self.__loader)
                        self.__loader = nil
                    end
                end
            )
        end
        self:onUnload()
        self:uninstall()
        -- 反注册界面事件
        if self.rec_events then
            for _evt, tab in pairs(self.rec_events) do
                for _path, is_reg in pairs(tab) do
                    if is_reg then
                        self:clearScript(_evt, _path)
                    end
                end
            end
        end

        self.rec_events = nil
    end, 

    events = {
        getter = function (self)
            local events = self._events
            if events == nil then 
                events = require("base.events").new()
                self._events = events
            end
            return events
        end, 
    }, 
    
    dispatch = function (self, evt_name, ...)
        local events = self._events
        if events then 
            events:dispatch(evt_name, self, ...)
        end
    end, 
    
    -- it returns the handler entry id
    on = function (self, evt_name, handler)
        return self.events:add(evt_name, handler)
    end, 

    -- 可以传入 handle_id 或者 handler 本身
    off = function (self, evt_name, handler)
        return self.events:remove(evt_name, handler)
    end, 

    -- 构造完成时
    onInit = function (self)
    end, 

    -- 资源加载完成时
    onLoaded = function (self)
    end, 

    -- onLoaded

    -- 界面销毁时
    onUnload = function (self)
    end, 

    -- 设置文本组件的文本内容
    setText = function (self, name, str)
        local text = self:findText(name)
        if text then
            text.text = str 
        end
        return text
    end, 

    -- 设置name按钮文本
    setButtonText = function (self, name, str)
        local btn = self:findButton(name)
        if btn then
            btn.text = str 
        end
        return btn
    end, 

    setSprite = function (self, name, assetPath)
        --print(name, assetPath)
        local image = self:findImage(name)
        if image then
            image:SetSprite(assetPath) 
        end
        return image
    end, 

    findTransform = function (self, name)
        if self.__loader then 
            return self.__loader:Find(name)
        end
    end, 

    find = function (self, name)
        return self:_getWrapper(name, UnityEngine.Transform)
    end, 

    findSlider = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.Slider)
    end, 

    findButton = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.Button)
    end, 

    findImage = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.Image)
    end, 

    findText = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.Text)
    end, 

    findInputField = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.InputField)
    end, 

    findDropdown = function (self, name)
        return self:_getWrapper(name, UnityEngine.UI.Dropdown)
    end, 

    -- donot_wrap: 不产生lua包装器，直接将c#对象存入
    _getWrapper = function (self, path, type_cls, donot_wrap)
        -- print("_getWrapper", path, type_cls)
        local transform_wrap = self._transforms[path]

        if not transform_wrap then
            local result = self:findTransform(path)
            if result then 
                transform_wrap = { result }
            else
                transform_wrap = { "nil" }
                -- logger.error("transform", path, "[", self.transform, "] not exist for type", type_cls)
            end
            self._transforms[path] = transform_wrap 
        end
        local transform = transform_wrap[1]
        if transform == "nil" then 
            return nil
        end

        local type_inst_wrap = transform_wrap[type_cls]
        if not type_inst_wrap then 
            if type_cls == UnityEngine.Transform then 
                type_inst_wrap = donot_wrap and transform or WidgetWrapper.new(transform)
            else
                local native_obj = transform:GetComponent(type_cls) or transform.gameObject:AddComponent(type_cls)

                if native_obj == nil then
                    type_inst_wrap = "nil"
                else 
                    if donot_wrap then 
                        type_inst_wrap = native_obj
                    else
                        if type_cls == UnityEngine.UI.Button then
                            type_inst_wrap = ButtonWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.InputField then
                            type_inst_wrap = InputFieldWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.Dropdown then
                            type_inst_wrap = DropdownWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.Image then 
                            type_inst_wrap = ImageWrapper.new(native_obj)
                            -- print(native_obj, type_inst_wrap)
                        else
                            type_inst_wrap = WidgetWrapper.new(native_obj)
                            -- print("wrap", path, native_obj)
                        end 
                    end
                end
            end
            transform_wrap[type_cls] = type_inst_wrap
        end
        --print("GetWrapper", path, type_cls, type_inst_wrap)
        return type_inst_wrap == "nil" and nil or type_inst_wrap 
    end,

    -- 备份一下  有些问题 
    --[[
    -- donot_wrap: 不产生lua包装器，直接将c#对象存入
    _getWrapper = function (self, path, type_cls, donot_wrap)
        local transform_wrap = self._transforms[path]

        if not transform_wrap then
            local result = self:findTransform(path)
            if result then 
                transform_wrap = { result }
            else
                transform_wrap = { "nil" }
                -- logger.error("transform", path, "[", self.transform, "] not exist for type", type_cls)
            end
            self._transforms[path] = transform_wrap 
        end
        local transform = transform_wrap[1]
        if transform == "nil" then 
            return nil
        end

        local type_inst_wrap = transform_wrap[type_cls]
        if not type_inst_wrap then 
            if type_cls == UnityEngine.Transform then 
                type_inst_wrap = donot_wrap and transform or WidgetWrapper.new(transform)
            else
                local native_obj = transform:GetComponent(type_cls) or transform.gameObject:AddComponent(type_cls)

                if native_obj == nil then
                    type_inst_wrap = "nil"
                else 
                    if donot_wrap then 
                        type_inst_wrap = native_obj
                    else
                        if type_cls == UnityEngine.UI.Button then
                            type_inst_wrap = ButtonWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.InputField then
                            type_inst_wrap = InputFieldWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.Dropdown then
                            type_inst_wrap = DropdownWrapper.new(native_obj)
                        elseif type_cls == UnityEngine.UI.Image then 
                            type_inst_wrap = ImageWrapper.new(native_obj)
                            -- print(native_obj, type_inst_wrap)
                        else
                            type_inst_wrap = WidgetWrapper.new(native_obj)
                            -- print("wrap", path, native_obj)
                        end 
                    end
                end
            end
            transform_wrap[type_cls] = type_inst_wrap
        end
        --print("GetWrapper", path, type_cls, type_inst_wrap)
        return type_inst_wrap == "nil" and nil or type_inst_wrap 
    end,
    ]]
    clearScript = function (self, evt, name)
        if evt == "click" then -- 单击 (必须是 Button 组件)
            return self:UnregisterClick(name)
        elseif evt == "drag" then  -- 拖拽
            return self:UnregisterDrag(name)
        elseif evt == "drop" then  -- 拖拽放下
            return self:UnregisterDrop(name)
        elseif evt == "updown" then -- 按下/抬起
            return self:UnregisterUpDown(name)
        elseif evt == "single" then 
            -- 单击, 与 click 的区别在于有 eventData 参数, 可以判定出drag状态下产生的单击事件
            -- 另外 click 目标对象必须是 Button 组件, single 可以为任意组件
            return self:UnregisterSingleClick(name)
        elseif evt == "double" then  -- 双击
            return self:UnregisterDoubleClick(name)
        elseif evt == "hover" then  -- 悬停(enter/exit)
            return self:UnregisterHover(name)
        elseif evt == "scroll" then 
            return self:UnregisterScroll(name)
        end
        return nil
    end,

    setScript = function (self, evt, name, callback)
        if evt == "click" then -- 单击 (必须是 Button 组件)
            return self:RegisterClick(name, callback)
        elseif evt == "drag" then  -- 拖拽
            return self:RegisterDrag(name, callback)
        elseif evt == "drop" then  -- 拖拽放下
            return self:RegisterDrop(name, callback)
        elseif evt == "updown" then -- 按下/抬起
            return self:RegisterUpDown(name, callback)
        elseif evt == "single" then 
            -- 单击, 与 click 的区别在于有 eventData 参数, 可以判定出drag状态下产生的单击事件
            -- 另外 click 目标对象必须是 Button 组件, single 可以为任意组件
            return self:RegisterSingleClick(name, callback)
        elseif evt == "double" then  -- 双击
            return self:RegisterDoubleClick(name, callback)
        elseif evt == "hover" then  -- 悬停(enter/exit)
            return self:RegisterHover(name, callback)
        elseif evt == "scroll" then 
            return self:RegisterScroll(name, callback)
        end
        return nil
    end,

    isLoaded = function(self)
        return self.is_loaded or false
    end,

    setButtonCD = function(self, name, cd_time)
        local trans = self.transform:Find(name)
        if trans then
            local btn = trans.gameObject:GetComponent(UnityEngine.UI.Button)
            if btn then
                btn.interactable = false
                app:startCoroutine(
                    function()
                        coroutine.yield(WaitForSeconds(cd_time or 0.5))
                        if self.gameObject and trans and btn then
                            btn.interactable = true
                        end
                    end
                )
            else
                log.error("Failed to find button:"..name)
            end
        else
            log.error("No :"..name)
        end
    end,

    -- 添加到命令缓存
    addCommandBuff = function(self, func)
        if func then
            self.command_buff = self.command_buff or {}
            table.insert(self.command_buff, func)
            if self:isLoaded() then
                self:excuteCommand()
            end
        end
    end,

    -- 执行命令缓存
    excuteCommand = function(self)
        --[[
        local canvas = self.transform:GetComponent(UnityEngine.Canvas)
        if canvas then
            local c = self.transform:AddComponent(UnityEngine.Canvas)
            c.overrideSorting = true
            c.sortingOrder = frame_manager:getSortingOrder(self.name)
            c.sortingLayerName = "UI"
            c.additionalShaderChannels = UnityEngine.AdditionalCanvasShaderChannels.TexCoord1;
            self.transform:AddComponent(UnityEngine.UI.GraphicRaycaster)
        end
        ]]

        if self.command_buff then
            for __, func in ipairs(self.command_buff) do
                if type(func) == "function" then
                    func()
                end
            end
            self.command_buff = nil
        end

        if frame_manager then
            frame_manager.events:dispatch("LOADING_COMPLETED", self.name)
        end

        
    end,

    uninstall = function(self)
        if frame_manager then
            frame_manager.events:dispatch("UNINSTALL_COMPLETED", self.name)
        end
    end,
}