local log = logger.get("Game", "Actor State skill")
local super = require('RTS.actor_states.actor_state_base')
local bullet_class = require('RTS.rts_bullet')
local actor_types = require('RTS.enums').actor_types
local bt_states = require('bt.node').states
local skill_hit_callbacks = require('RTS.skill_lib._hit_callback')

-- 技能状态
return class {
    super = super,
    ctor = function(self, ...)
        super.ctor(self, ...)
        self.all_hit_events = {}    -- 每个 cast inst id 对应的受击列表 cast_inst_id - hit event list    
        self.hit_event_indices = {} -- cast_inst_id - hit events index 记录每个 cast_inst_id 对应 hit_events 的当前索引
    end,

    onEnter = function(self, args)

        self.skill_id = args.id

        --[[
        if self.skill_id ~= 20004 then
            log.info("##### skill id:"..self.skill_id)
        end]]

        -- 备份原来的目标敌人
        self.target_actor = args.ta
        if self.target_actor then
            self.orig_target = self.owner.target_actor
            self.owner:setTargetActor(self.target_actor)
        else 
            self.target_actor = self.owner.target_actor
        end
        self.owner.skill_hit_callback = skill_hit_callbacks[self.skill_id]
        if self.owner.skill_hit_callback then
            self.owner.skill_hit_count = 0
            self.owner.hit_record = nil
        end

        self.skill_position = args.p   -- 用户选择的位置
        self.skill_rotation = args.r   -- 必杀技用户选择的方向
        self.skill_cfg = config.skill_skills[self.skill_id]
        if not self.skill_cfg then
            log.error('Failed to find skill cfg:'..tostring(self.skill_id)..', actor id:'..self.owner.id)
            self.skill_cfg = config.skill_skills[1]
        end

        self.owner.skill_cfg = self.skill_cfg   -- 将当前技能记录到角色上
        local bt_id = self.skill_cfg.bt_id or 1 -- 行为树 id

         self.is_finished = false
        self.state_elapsed = 0.0

        -- 创建技能行为树
        local bt_fun = require('RTS.skill_lib._skill_bt')[bt_id]
        if not bt_fun then 
            log.error('failed to find skill_bt:'..tostring(bt_id)..'in skill:'..self.skill_id)
            bt_fun = require('RTS.skill_lib._skill_bt')[1]
        end

        self.bt = bt_fun(self.owner)

        if self.bt == nil then
            log.error('failed to find skill bt:'..tostring(self.skill_id))
        end

        self:stopMove()
        self.casting_complete = false

        --[[ 战报播放相关 ]]
        if args.cast_inst_id then
            self.owner.cast_inst_id = args.cast_inst_id
            self.hit_event_indices[args.cast_inst_id] = 1
         else    
            self.owner.cast_inst_id = battle_helper:genSkillCastInstID()  -- 施放实例 id  记录在 actor 上
        end

        -- 预先确定的击中事件
        if args.hit_events and args.cast_inst_id  then
            self.all_hit_events[args.cast_inst_id] = args.hit_events
        end

        self.pos_list = args.pos_list   -- 战报中的位置列表
        self.pos_idx = 1
        
    end,

    -- 设置目标
    setTargetActor = function(self, target_actor)
        self.target_actor = target_actor
    end,

    -- 从 hit_events 中弹出顶部值, 技能的伤害可能跨技能时间，要保存到外部比如子弹上
    popHitEvent = function(self, cast_inst_id, target_id)
            local hit_events = self.all_hit_events[cast_inst_id]
            if not hit_events then
                return nil
            end
            for k,v in ipairs(hit_events) do
                if v.target_id == target_id then
                    table.remove(hit_events, k)
                    return v
                end
            end

           --log.error("Failed to find hit event cast_inst_id:"..cast_inst_id..", target_id:"..target_id)
            return nil -- {is_hit=false, is_crit=false, hp=0, evt_type=1}

    end,

    onUpdate = function(self, dt)
        self.state_elapsed = self.state_elapsed + dt
        if self.state_elapsed > 30 then
            log.info('state_elapsed over 30!! actor:'..self.owner.id)
        end

        if not self.is_finished then 
            local rst = self.bt:update(dt, 2)
            if rst == bt_states.SUCCESS then 
                self:finished()
            elseif rst == bt_states.FAILURE then
                log.info("# failure!!")
                self:finished()
            end
        end
    end,

    canChangeTo = function(self, next_state_name, args) 

        --[[
        if self.owner.team == 1 then
            log.info("### next:"..next_state_name)
            log.info(debug.traceback())
        end]]

        if self.is_finished or next_state_name == 'escape' or next_state_name == "die" or next_state_name == "hit_fly" then
            return true
        elseif self.owner.is_in_death_skill then
            return false
        elseif next_state_name == 'preskill' then
            return true
        elseif next_state_name == 'skill' then   -- 必杀技，
            local cfg = config.skill_skills[args.id]
            if cfg.skill_type == 1 and self.skill_cfg.skill_type == 0 then
                return true
            end
        elseif next_state_name == 'hit_back' and not self.casting_complete then
            return true
        end

        --log.info('change state failed:'..next_state_name)
        return false
    end,

    finished = function(self)
        self.is_finished = true
        if self.owner.is_alive then
            self.owner:changeState('idle')
        else
            self.owner:changeState('die')
            self.owner.events:dispatch('actor_dead', self.owner)
        end
    end,

    setContext = function (self, ctx)
       self.ctx = ctx
       if self.child ~= nil then
        self.child:setContext(ctx)
       end
    end, 

    -- 重置
    reset = function(self)
        self.child:reset()
    end,

    onExit = function(self)
        self.skill_id = 0
        self.target_actor = nil
        self.skill_position = nil
        self.skill_rotation = nil
        self.owner.skill_cfg = nil

        if self.bt then
            self.bt:reset()
            self.bt = nil
        end

        -- 恢复原来的目标敌人
        if self.orig_target and self.orig_target:isAlive() then
            self.owner:setTargetActor(self.orig_target)
        end

    end,
}