-- AI 世界
local log = logger.get("Game", "AI World")
local agent_class = require('RTS.ai.agent')
local vec2 = require('base.vec2')
local actor_types = require('RTS.enums').actor_types

return class {
    ctor = function(self)
        self.agents = {}   
        self.actors = {}
        self.paused = true
        self.marks = {}
        self.actor_array = {}   -- actor 有序数组

        app:startCoroutine(
            function()
                local idx = 1
                while not self.is_destroyed do
                    if not self.paused then
                        -- 每帧同步有限数量的 actor 位置
                        if idx > #self.actor_array then
                            idx = 1
                        end
                        local actor = self.actor_array[idx]
                        if actor then
                            local agent = self.agents[actor.id]
                            if agent.removed then
                                self.actor_array[idx] = nil
                            else
                                agent:setPosition(actor:getPosition())
                            end
                        end

                        idx = idx + 1
                    end
                    coroutine.yield(nil)
                end
            end
        )
        ai_world = self
    end,

    -- 创建 actor 的 ai 代理
    createAI = function(self, actor)
        -- agent 已经存在
        if self.agents[actor.id] then
            log.error("AI agent already exist!, id:"..actor.id)
            log.error(debug.traceback())
            return
        end
        
        local agent = agent_class.new(self, actor.id, actor.ai_name, actor.team, actor.actor_type, 0.5, actor.attack_range, actor.min_attack_range, actor.safe_range)
        agent.priority_type = actor.cfg.type    -- agent 本身的优先级类型
        agent.select_enemy_type_id = actor.cfg.atk_target    -- 选择敌人类型 tank_basic_get_targets 表 idx
        agent.radius = actor.radius
        agent.attack_cd = actor:computeSkillCD()
        
        self.agents[agent.id] = agent
        self.actors[actor.id] = actor
        
        table.insert(self.actor_array, actor)

        if actor.owner then
            agent.owner = self.agents[actor.owner.id]
            --log.info('self. owner:'..actor.owner.id..tostring(agent.owner))
        end
        --if actor.team == 1 then
        --    log.info('Create agent:'..actor.id..","..actor.ai_name)
        --end


    end,

    -- 移除 actor 的 ai 代理
    removeAI = function(self, actor)
        local agent = self.agents[actor.id]
        if not agent then return end

        agent.removed = true
        --log.info('remove agent:'..actor.id)

        -- 找到所有以 agent 做为敌人的 ai，重置
        for _,v in pairs(self.agents) do
            if v.enemy == agent then
                -- 移除敌人，重置 ai

                self:clearAgentEnemy(v)
            end
        end
    end,

    resetAI = function(self, actor)
        local agent = self.agents[actor.id]
        if not agent then return end

        agent:reset()
        agent.enemy = nil
    end,

    update = function(self, dt)
        if self.paused or dt == 0.0 then
            return
        end

        for _, agent in pairs(self.agents) do

            local actor = self.actors[agent.id]
            
            -- 每帧更新 agent 的位置
            if agent.removed or not actor:isAlive() or actor:isEscape() then
                agent.is_alive = false
            else

                if not agent.is_sneak and actor.is_sneak then
                    self:onBeginSneak(agent)
                end
                agent.is_sneak = actor.is_sneak
            end
        end

        for _, agent in pairs(self.agents) do
            if agent.is_alive  then
                agent:update(dt)
            end
        end

    end,

    -- 如果敌人是 agent，将其清除
    onBeginSneak = function(self, enemy)
        for _, agent in pairs(self.agents) do
            -- 每帧更新 agent 的位置
            if agent.is_alive  and agent.enemy == enemy then
                self:clearAgentEnemy(agent)
            end
        end
    end,

    setPaused = function(self, paused)
        -- log.info('aiWorld SetPaused:'..tostring(paused))
        if self.paused and not paused then

            for _, agent in pairs(self.agents) do
                if agent.is_alive then
                    local actor = self.actors[agent.id]
                    agent:setPosition(actor:getPosition())
                end
            end
        end

        self.paused = paused
    end,

    -- 获取距离 pos 最近的 agent
    getNearestAgent = function(self, pos, team, target_actor_type)
        local min_d_enemy = nil
        local min_d = 99999999

        -- log.debug('getNearestAgent, x:'..pos[1]..', y:'..pos[2]..', team:'..team)
        -- 遍历所有 agent
        for _, agent in pairs(self.agents) do 
            repeat 
                if agent.team == team and agent:isAlive() and not agent.removed and not agent.is_sneak then
                    if target_actor_type and agent.actor_type ~= target_actor_type then
                        break -- 相当于 for 循环的 continue
                    end

                    local d = vec2.dist_sq(pos, agent.pos)
                    if d < min_d then
                        min_d = d
                        min_d_enemy = agent
                    end
                end
            until true
        end
        
        return min_d_enemy, min_d
    end,

    --[[
    -- 选择敌人的agent (回合制)
    selectEnemyAgent = function(self, self_agent)
        local enemy_agents, enemy_actors = self:selectEnemyByTurn(self_agent, nil)
        return enemy_agents
    end,

    -- 选择敌人的actor (回合制)
    selectEnemyActor = function(self, self_actor)
        local enemy_agents, enemy_actors = self:selectEnemyByTurn(nil, self_actor)
        return enemy_actors
    end,

    -- 选择敌人 (回合制)
    selectEnemyByTurn = function(self, self_agent, self_actor)

        if self_agent then
            self_actor = self.actors[self_agent.id]
        elseif self_actor then
            self_agent = self.agents[self_actor.id]
        end

        if self_agent == nil or self_actor == nil then
            log.error("选择敌人 self_agent or self_actor is nil")
            return
        end

        local enemy_actors = {}
        local enemy_agents = {}

        local cfg_id = self_actor:getSkillID()
        local skill_cfg = config.skill_skills[cfg_id]
        if skill_cfg then
            local id = skill_cfg.target
            id = 3 -- 测试
            print("测试 回合制选择敌人规则", id)

            if id == 1 then
            -- 1. 默认单体-battle_config里面

                local atk_cfg = config.battle_config_atk_configs[self_actor.slot_pos]
                local split_tb = battle_helper:split(atk_cfg.default_order)

                for k, slot_pos in ipairs(split_tb) do
                    local temp_actor = self:getActorBySlotPos(self_actor.team, slot_pos)
                    if temp_actor and temp_actor:isAlive() then
                        enemy_actors[temp_actor.id] = temp_actor
                        break
                    end
                end

            elseif id == 2 then
            -- 2. 后排单体-battle_config里面

                local atk_cfg = config.battle_config_atk_configs[self_actor.slot_pos]
                local split_tb = battle_helper:split(atk_cfg.type1)

                for k, slot_pos in ipairs(split_tb) do
                    local temp_actor = self:getActorBySlotPos(self_actor.team, slot_pos)
                    if temp_actor and temp_actor:isAlive() then
                        enemy_actors[temp_actor.id] = temp_actor
                        break
                    end
                end

            elseif id == 3 then

                for id, act in pairs(self.actors) do
                    if act and act:isAlive() and act.team ~= self_actor.team then
                        enemy_actors[id] = act
                    end
                end

            end
        else
            log.error('agent 找不到对应 skill_cfg !!!', cfg_id)    
        end

        for id, actor in pairs(enemy_actors) do
            enemy_agents[id] = self.agents[id]
        end

        return enemy_agents, enemy_actors
    end,]]

    -- 根据优先级和距离选择敌人
    selectEnemyByPriority = function(self, self_agent, only_attackable)
        local max_p_enemy = nil
        local max_p = 0

        local prioritys = config.tank_basic_get_targets[self_agent.select_enemy_type_id]
        if not prioritys then
            prioritys = config.tank_basic_get_targets[1]
        end

        -- 遍历所有 agent
        for _, agent in pairs(self.agents) do 
            if agent.team ~= self_agent.team and agent:isAlive() and not agent.removed and not agent.is_sneak then
                -- 计算优先级
                local p = prioritys[agent.priority_type]
                local d = vec2.dist(self_agent.pos, agent.pos)
                if d > 0.001 then
                    if not only_attackable or 
                        (d > self_agent.min_attack_range and d < self_agent.attack_range) then
                        p = p / d
                        if p > max_p then
                            max_p = p
                            max_p_enemy = agent
                        end
                    end
                end
            end
        end

       -- log.info("### agent select enemy:"..self_agent.id..", max_p"..max_p)
        return max_p_enemy
    end,

    -- 获取 agent 所有敌人
    getAllEnemy = function(self, self_agent, only_attackable)
        local all_enemy = {}
        for _, agent in pairs(self.agents) do 
            if agent.team ~= self_agent.team and agent:isAlive() and not agent.removed 
                and not agent.is_sneak and agent.actor_type == actor_types.VEHICLE then
                local d = vec2.dist(self_agent.pos, agent.pos)
                agent.temp_dist = d -- 临时保存到 agent 的距离

                if only_attackable then
                    if not only_attackable or 
                        (d > self_agent.min_attack_range and d < self_agent.attack_range) then
                        table.insert(all_enemy, agent)
                    end
                else
                    table.insert(all_enemy, agent)
                end
            end
        end
        return all_enemy
    end,

    -- 随机选择几个
    selectMultipleEnemyRandomly = function(self, agent, num, only_attackable)
        if not agent then
            return nil
        end
        local all_enemy = self:getAllEnemy(agent, only_attackable)

        local selected = {}
        for i=1,num do
            if #all_enemy == 0 then
                break
            end

            local idx = math.random(#all_enemy)
            table.insert(selected, all_enemy[idx])
            table.remove(all_enemy, idx)
        end

        if #selected > 0 then
            return selected
        else
            return nil
        end
    end,

    -- 距离最近的几个
    selectMultipleEnemyByRange = function(self, agent, num, only_attackable)
        if not agent then
            return nil
        end
        local all_enemy = self:getAllEnemy(agent, only_attackable)

        -- 根据距离排序
        table.sort(all_enemy, function(a, b)
            if a.temp_dist < b.temp_dist then
                return true
            elseif a.temp_dist == b.temp_dist then
                return false
            else
                return false
            end
        end)

        local selected = {}
        for i=1,num do
            if #all_enemy < i then
                break
            end

            table.insert(selected, all_enemy[i])
        end

        if #selected > 0 then
            return selected
        else
            return nil
        end
    end,

    -- 根据槽位获取actor
    getActorBySlotPos = function(self, team, slot_pos)
        if type(slot_pos) == "string" then
            slot_pos = tonumber(slot_pos)
        end
        local rst = nil
        for __, actor in pairs(self.actors) do
            if actor.slot_pos == slot_pos and actor.team ~= team then
                rst = actor
                break
            end
        end
        return rst
    end,

    -- 目标敌人发生改变 
    enemyChanged = function(self, agent, enemy)
        local actor = self.actors[agent.id]
        if enemy then
            local enemy_actor = self.actors[enemy.id]
            actor:ai_setTargetActor(enemy_actor)
        else
            actor:ai_setTargetActor(nil)
        end
    end,

    --[[
    -- 回合制敌人发生变化
    enemyChangedTb = function(self, agent, tb_enemys)
        local actor = self.actors[agent.id]
        local rst = {}
        for id, agt in pairs(tb_enemys) do
            rst[id] = self.actors[agt.id]
        end
        actor:ai_setTargetActorTb(rst)
    end,]]

    -- 跟随宿主坦克并使用技能
    followOwnerTankAndAttack = function(self, agent)
        local actor = self.actors[agent.id]
        if actor.owner then
            local owner_agent = self.agents[actor.owner.id]
            agent.enemy = owner_agent
            actor:ai_setTargetActor(actor.owner)
            actor:ai_followAndAttack(actor.owner)
        else
            log.error('followOwnerTank, no owner!!')
        end
         
    end,

    -- agent 进入 idle 状态
    agentIdle = function(self, agent)
        local actor = self.actors[agent.id]
        actor:changeState('idle')
    end,

    -- 移动到敌人，进入射程后开始攻击
    agentMoveToEnemyAndAttack = function (self, agent)
        local actor = self.actors[agent.id]
        local enemy = self.actors[agent.enemy.id]

        if actor and enemy then
            actor:ai_moveToAndAttack(enemy)
        end
    end,

    -- 脱离战场
    --[[
    agentEscape = function(self, agent)
        --log.info('Agent Escape!!')
        local actor = self.actors[agent.id]
        actor:ai_escape({80, 80})   -- 场景在 x, z 方向的 extend
    end,]]

    clearAgentEnemy = function(self, agent)
        agent.enemy = nil
        agent:reset()
        local actor = self.actors[agent.id]
        actor:clearTargetActor()
    end,

    --
    -- 向敌人移动
    agentMoveToEnemy = function (self, agent)
        local actor = self.actors[agent.id]
        local enemy = self.actors[agent.enemy.id]
     
        if actor and enemy then
            actor:ai_moveToTarget(enemy)
        end
    end,

    -- 向敌人反向移动
    agentMoveAwayFromEnemy = function(self, agent)
        local actor = self.actors[agent.id]
        local enemy = self.actors[agent.enemy.id]
     
        if actor and enemy then
            local pos = actor:getPosition()
            local to = pos - enemy:getPosition()
            to:Normalize()
            pos = pos + to*10
            actor:ai_moveToPos(pos)
        end
    end,

    -- 转向面向敌人 
    agentFacingTarget = function(self, agent)
        if not agent.enemy then
            return
        end

        local actor = self.actors[agent.id]
        local enemy = self.actors[agent.enemy.id]
        actor:changeState("idle")
        actor.cs_inst:FacingTargetOnTheSpot(enemy:getPosition())
    end,

    -- 攻击敌人
    agentAttackEnemy = function (self, agent, enemy)
        local actor = self.actors[agent.id]
        local enemy_actor = self.actors[enemy.id]

        if actor and enemy_actor then
            return actor:ai_attackTarget(enemy_actor)
        else
            log.error('failed to find actor or enemy:'..tostring(agent)..','..tostring(enemy))
            return false
        end
    end,

    -- 回合制 攻击敌人
    agentAttackEnemyTb = function(self, agent, enemy_agents)
        local actor = self.actors[agent.id]

        local enemy_actors = {}
        for id, agt in pairs(enemy_agents) do
            enemy_actors[id] = self.actors[id]
        end

        if actor and enemy_actors then
            return actor:ai_attackTargetTb(enemy_actors)
        else
            log.error("回合制 攻击敌人 失败 ！！！")
            return false
        end       
    end,

    -- 施放技能
    agentCastSkill = function(self, agent, enemy, skill_id)
        local actor = self.actors[agent.id]
        local enemy_actor = self.actors[enemy.id]

        return actor:ai_castUltimateSkill(enemy_actor, skill_id)
    end,

    -- 释放开场技能
    agentCastOpeningskill = function(self, agent, force)
        local actor = self.actors[agent.id]
        actor:ai_castOpeningSkill(force);
    end,

    -- 获取 mark 容器
    getMarkTab = function(self, agent, is_global)
        if is_global then
            return self.marks
        else
            local actor = self.actors[agent.id]
            if actor then
                return actor.marks
            else
                log.error('Failed to find actor:'..agent.id)
                return nil
            end
        end
    end,

    --[[标记接口]]
    -- 添加一个标记
    agentAddMark = function(self, agent, mark_name, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)

        if not mark_tab[mark_name] then
            mark_tab[mark_name] = 0
        else
            log.error('mark already exist:'..mark_name)
        end
        return true
    end,

    isMarkExist = function(self, agent, mark_name, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        return mark_tab[mark_name]
    end,

    agentDelMark = function(self, agent, mark_name, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        return mark_tab[mark_name] == nil
    end,

    agentIncMark = function(self, agent, mark_name, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            mark_tab[mark_name] = mark_tab[mark_name] + 1
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    agentDecMark = function(self, agent, mark_name, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            mark_tab[mark_name] = mark_tab[mark_name] - 1
            return true
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    agentSetMark = function(self, agent, mark_name, value, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            mark_tab[mark_name] = value
            return true
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    isMarkEqual = function(self, agent, mark_name, value, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            return mark_tab[mark_name] == value
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    isMarkGreatThan = function(self, agent, mark_name, value, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            return mark_tab[mark_name] > value
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    isMarkLessThan = function(self, agent, mark_name, value, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        if mark_tab[mark_name] then
            return mark_tab[mark_name] < value
        else
            log.error('mark not exist:'..mark_name)
            return false
        end
    end,

    printMarks = function(self, agent, is_global)
        local mark_tab = self:getMarkTab(agent, is_global)
        for k,v in pairs(mark_tab) do
            -- log.info('mark['..tostring(k)..'] = '..tostring(v))
        end
    end,

    addToCameraView = function(self, agent)
        local actor = self.actors[agent.id]
        battle_helper:addToCameraView(actor)
    end,

    removeFromCameraView = function(self, agent)
        local actor = self.actors[agent.id]
        battle_helper:removeFromCameraView(actor)
    end,

    -- ai 执行结束
    finished = function(self, agent)
        agent.removed = true
    end,

    destroy = function(self)
        self.is_destroyed = true
    end,
}