local log = logger.get("Game", "RTSActor")
local trace = require('debug.trace')
local bullet_class = require('RTS.rts_bullet')
local list = require('base.list')
local actor_types = require('RTS.enums').actor_types
local super = require('common.state_machine')

-- 角色基类,只定义子类必须要实现的一些函数
return class{ 
    super = super,
    extensions = { "RTS/rts_actor_base+ai.lua", "RTS/rts_actor_base+skill.lua"--[[, "RTS/rts_actor_base+tb.lua"]]}, 

    ctor = function(self, team, id)
        super.ctor(self)

        init_props = nil        -- 服务器下发的属性信息 MParPropIdValue 数组
        init_weapon = nil      -- 服务器下发的武器信息 整数数组
        self.team = team
        self.id = id
        self.radius = 1.0
        self.height = 1.0
        
        -- states
        self.is_initialized = false  -- 是否初始化完成
        self.is_alive = true    -- 是否活着
        self.is_moving = false  -- 是否在移动

        self.attack_elapsed = 99.0
        self.attack_range = 40
        self.attack_range_sq = self.attack_range * self.attack_range
        
        self.hp = 0
        self.energy = 0
        self.shield = 0

        self.anim_names =       -- 动画名称配置 
        {
            idle = nil,
            move = nil,
            stop = nil,
            attack = nil,
            die = nil,
        }    -- 

        self.props = nil -- 角色属性包，actor_props 的子类
        self.buff_handler = nil --  buff 处理器

        -- sub ai
        self.group_target_actor = nil   -- soldier group 目标
        self.target_actor = nil         -- 单个目标，可能是坦克或一个士兵
        self.should_attack = false      -- 是否要攻击目标
        --self.check_elapsed = math.random() * 0.5    -- [0, 0.5]
        --self.check_time = math.random()*0.5 + 0.2   -- 多久检查一次敌人是否在射程内[0.2, 0.7] 

        -- 一帧常用数据缓存，防止一帧内重复调用 c# 接口
        self.cache_position = nil
        self.cache_rotation = nil
        self.cache_is_facing_target = nil

        self.marks = {}
        self.last_hit_actor = nil
        self.init_buffs = {}

        self.damage_text = {}   -- 头上飞的伤害跳字
    end,

    -- 初始化, 子类实现
    init = function(self)

    end,

    cs_init = function(self)
        if self.cs_inst then
            self.cs_inst:Init(self.team, self.id, self.group_id or 0, true)
            self.cs_inst.enabled = true    
            self.cs_inst:SetMoveParam(self.speed, self.acceleration, 90, 1.0)
            self.cs_inst:SetAvoidancePriority(self.cfg.avoid_priority)
        end
    end,

    preUpdate = function(self, dt)
        self.cache_position = false
        self.cache_rotation = false
        self.cache_is_facing_target = false
    end,

    update = function(self, dt)
        if not self.is_initialized then
            return
        end

        super.update(self, dt)
        if self.buff_handler then
            self.buff_handler:update(dt)
        end

        self.attack_elapsed = self.attack_elapsed + dt
    end,

    isAlive = function(self)
        return self.is_alive
        --log.error('isAlive 子类应该实现本函数!!')
    end,

    -- 是否在溃逃
    isEscape = function(self)
        return self.is_escape
    end,

    getPosition = function (self)
        if self.transform == nil then
            log.error(debug.traceback())
            return Vector3.zero
        end

        if not self.cache_position then
            self.cache_position = self.transform.position
        end
        return self.cache_position
    end,

    setPosition = function(self, pos)
        self.transform.position = pos
    end,

    getRotation = function(self)
        if not self.cache_rotation then
            self.cache_rotation = self.transform.rotation
        end
        return self.cache_rotation
    end,

    setRotation = function(self, r)
        self.cs_inst:SetRotation(r)
    end,

    -- 获取角色的半径
    getRadius = function(self)
        return self.radius
    end,

    -- 获取敌对 team
    getEnemyTeam = function(self)
        if self.team == 0 then
            return 1
        else
            return 0
        end
    end,

    -- 获取动画后缀
    getAnimPostfix = function(self)
        return nil
    end,

    -- 获取普通攻击技能 id，子类应该重写
    getCommonAtkID = function(self)
        return 1
    end,

    -- 获取技能 id, 子类应当重写
    getSkillID = function(self)
        return 100
    end,

    -- 释放技能
    castSkill = function(self, skill_id, target)
        if self:isInSkillCD(skill_id) then
            return false
        end

        if self:changeState('skill', {id = skill_id, ta = target}) then
            self.attack_elapsed = 0.0
        end
        return true
    end,

    -- 是否在攻击 cd 中 
    isInSkillCD = function(self, skill_id)
        -- 目前简单判断
        return self.attack_elapsed < self:computeSkillCD(skill_id)
    end,

    -- 计算当前技能 cd
    computeSkillCD = function(self, skill_id)
        local id = skill_id or self:getCommonAtkID()
        local cfg = config.skill_skills[id]   -- 技能配置
        if not cfg then
            log.error("Failed to find skill id:"..tostring(id))
            return 9999
        end
        if cfg.skill_type == 1 or cfg.skill_cd == 0 then
            -- 不是普通攻击，直接用技能 cd
            return cfg.skill_cd
        end

        if not self.props then
            return 9999
        end

        local atk_speed = self.props:getProp(ATK_SPEED_IDX)
        local atk_speed_factor = self.props:getProp(MAX_HP_FACTOR_IDX)

        if atk_speed == 0 and atk_speed_factor == 0 then
            -- 无属性加成
            return cfg.skill_cd
        else
            local skill_speed = 1.0 / cfg.skill_cd
            skill_speed = (skill_speed + atk_speed) * (1.0 + atk_speed_factor)

            return 1.0 / skill_speed
        end
    end,

    getTargetActor = function(self)
        return self.target_actor
    end,

    -- 当前目标是否在攻击范围内
    isTargetInAttackRange = function(self)
        --log.info(self.name..', attack range sq:'..self.attack_range_sq)
        if self.target_actor and self.target_actor:isAlive() then
            return Vector3.Distance(self:getPosition(), self.target_actor:getPosition()) <= self.attack_range + self.target_actor.radius * 0.8
        else
            return false
        end
    end,

    -- 绑定 obj 到射击点
    bindToShootPoint = function (self, obj, rot)
        return self.cs_inst:BindToShoot(obj, Vector3(rot[1], rot[2], rot[3]))
    end, 

    -- 绑定 obj 到头顶
    bindToHeadPoint =  function (self, obj, rot)
        local rot = rot or {0,0,0}
        return self.cs_inst:BindToHead(obj, Vector3(rot[1], rot[2], rot[3]))
    end, 

    -- 绑定 obj 到任意 bone
    bind = function (self, bone, obj, rot)
        local rot = rot or {0,0,0}
        return self.cs_inst:Bind(bone, obj, Vector3(rot[1], rot[2], rot[3]))
    end, 

    getActorType = function(self)
        return self.actor_type
    end,

    setTargetActor = function(self, target_actor)

        local orig = self.target_actor
        --log.info(self.id.." setTargetActor:"..tostring(target_actor))
        --log.info(debug.traceback())

        -- 如果是 soldier group, 选择最近的一个做为目标
        if not target_actor then
            self:clearTargetActor()
        elseif target_actor:getActorType() == actor_types.SOLDIER_GROUP then
            if target_actor ~= self.group_target_actor then
                self.group_target_actor = target_actor
                self.target_actor = battle_helper:selectNearestFromGroup(self:getPosition(), target_actor)
            end
        else
            self.group_target_actor = nil
            self.target_actor = target_actor
        end

        if self.target_actor then
            self.cs_inst:SetTargetActor(self.target_actor.transform)
        else
            self.cs_inst:SetTargetActor(nil)
        end
        self.cache_is_facing_target = nil
        
        if self.main_weapon and target_actor then
            self.main_weapon:facingTarget(target_actor.transform)
        end
    end,

    -- 清除当前目标敌人
    clearTargetActor = function(self)

        if self:isAlive() then
            self.target_actor = nil
            self.group_target_actor = nil
            self:changeState('idle')
            battle_helper:resetAI(self)
            self.cs_inst:SetTargetActor(nil)

        end
    end,

    -- 设置移动速度
    setSpeed = function(self, speed)
        self.speed = speed
        self.acceleration = speed
        self.cs_inst:SetMoveParam(self.speed, self.acceleration, 90, 1.0)
    end,

    -- 在 time 时间后deactive, 死亡或者脱离战场后调用
    hideGameObject = function(self, time)
        if time then
            app:startCoroutine(function()
                coroutine.yield(WaitForSeconds(time))
                self.gameObject:SetActive(false)
            end)
        else
            self.gameObject:SetActive(false)
        end
    end,

    doChangeState = function(self, name, args)
        local curr_state = super.doChangeState(self, name, args)
        return curr_state
    end,

    -- 血量变化
    addHP = function (self, _hp)
       
        local hp = _hp
        if hp < 0 and self.shield > 0 then
            -- 优先减护盾
            self.shield = self.shield + hp
            if self.shield < 0 then
                hp = self.shield
                self.shield = 0
            else
                hp = 0
            end
        end

        self.hp = self.hp + hp
        if self.hp > self.max_hp then
            self.hp = self.max_hp
        end

        if self.hp <= 0 then
            --self.is_alive = false

            if battle_helper:isReplay() then
                --log.info("actor hp < 0 :"..self.hp)
            else
                self.is_alive = false
                -- 如果有亡语技能，先触发亡语
                if self.cfg.death_skill > 0 then
                    if not self.is_in_death_skill then
                        self:changeStateImmediate('skill', {id=self.cfg.death_skill, ta=self.target_actor})
                        self.is_in_death_skill = true
                    end
                else
                    self:changeState('die')
                    self.events:dispatch('actor_dead', self)
                end
            end
        end

        return self.hp
    end,

    kill = function(self)
        self.hp = 0
        self.is_alive = false
        self:doChangeState('die')
        self.events:dispatch('actor_dead', self)
    end,

    addDamageText = function(self, obj)
        table.insert(self.damage_text, obj)
    end,

    removeDamageText = function(self, obj)
        for k,v in ipairs(self.damage_text) do
            if v == obj then
                table.remove(self.damage_text, k)
                break
            end
        end
    end,

    destroy = function(self)
        if self.main_weapon then
            self.main_weapon:destroy()
            self.main_weapon = nil
        else
        end

        if self.trailer_body_go then
            self.trailer_body_go:destroy()
        end
    
        if self.head_info_go then
            self.head_info_go:destroy()
        end

        if self.blob_shadow then
            self.blob_shadow:destroy()
        end

        if self.gameObject then
            self.gameObject:destroy()
        end
    end,

}