-- 引导管理器
local log = logger.get("Game", "Guide")
require("guide.guide_helper")
require("config.guide")

local guide_manager_class = class{
    ctor = function(self)
        self.complete_guide_id = nil  -- 已完成的顺序引导
        self.random_guide_complete = nil -- 随机引导完成情况 [id-true] 标识完成

        self.running_guide_id = 0   -- 正在执行的引导

        self.waiting_tap_count_down = 0

        -- 找出所有随机引导
        self.random_guides = {}
        config.guide_guides:foreach(
            function(k,v)
                if k > 10000 then
                    self.random_guides[k] = v
                end
            end
        )
    end,

    -- 判断引导是否已经完成
    isGuideComplete = function(self, id)
        if id < 10000 then
            return id <= self.complete_guide_id
        else
            return self.random_guide_complete[id]
        end
    end,

    isGuideRunning = function(self)
        return self.running_guide_id > 0
    end,

    -- 尝试开始引导
    tryStartGuide = function(self)
        --return

        if not self.complete_guide_id then
            -- 还没有初始化
            return
        end

        if self:isGuideRunning() then
            return false
        end

        -- 尝试开启顺序引导
        local guide_cfg = self:getNextGuide()
        if guide_cfg then
            -- 尝试 guide_cfg 是否能执行
            if self:isFitCondition(guide_cfg) then
                if self:_doGuide(guide_cfg.id) then
                    return true
                end
            end
        end

        -- 尝试开启随机引导
        for k,v in pairs(self.random_guides) do
            if not self.random_guide_complete[k] then
                if self:isFitCondition(v) then
                    if self:_doGuide(v.id) then
                        return true
                    end
                end
            end
        end
       
        return false
    end,

    -- 判断引导是否满足执行条件
    isFitCondition = function(self, guide_cfg)
        if not guide_cfg then
            return false
        end

        -- 引导触发界面是否开启
        if not frame_manager:isDisplay(guide_cfg.trigger_frame) then
            return false
        end

        -- 个人基地等级
        if home_info:getMainEntityLv() < guide_cfg.home_level then
            return false
        end

        return true
    end,

    getNextGuide = function(self)
        local id = self.complete_guide_id + 1
        return config.guide_guides[id]

    end,

    -- 执行一个引导
    _doGuide = function(self, id)
        --return false

        local guide_script = require("guide/guide_script/guide_"..tostring(id))
        if not guide_script then
            log.error("Failed to find guide:"..tostring(id))
            return false
        end

        -- 引导执行条件
        if guide_script.condition then
            local can_run = guide_script.condition()
            if not can_run then
                -- 不满脚本足执行条件
                return false
            end
        end

        self.running_guide_id = id

        self.guide_script = guide_script
        app:startCoroutine(
            function()
                game_director.pool:wait_prepare("Assets/Prefab/UI/TapHere.prefab", 5)
                -- 以协程方式运行引导脚本
                self.coroutine_id = app:startCoroutine(guide_script.proc)
            end
        )

        return true
    end,

    -- 引导被打断
    onGuideBreak = function(self)
        if self.guide_script then
            if self.guide_script.onBreak then
                self.guide_script.onBreak(self.running_guide_id)
            end

            -- 把该关的都关掉
            guide_helper:disableScreenMask()
            guide_helper:removeTapHere()
            guide_helper:guideSpriteHide()

            -- 停止协程
            if self.coroutine_id then
                app:stopCoroutine(self.coroutine_id)
                self.coroutine_id = nil
            end
            self.running_guide_id = -1
        end
    end,

    -- 引导完成
    onGuideComplete = function(self, id, do_not_try_next)
        log.info("### onGuideComplete:"..id)

        if id < 10000 then
            if self.complete_guide_id < id then
                self.complete_guide_id = id
            else
                log.error("Error in guide!, id:"..tostring(id))
            end
        else
            self.random_guide_complete[id] = true
            app:logEvent("guide_complete_"..tostring(id), tostring(id))
        end
        self:saveGuideData()

        guide_helper:showSkipButtonCallback(false)
        
        self.running_guide_id = -1

        if not do_not_try_next then
            self:tryStartGuide()
        end
    end,

    update = function(self, dt)
        if self:isGuideRunning() and self.waiting_tap_count_down > 0 then
            self.waiting_tap_count_down = self.waiting_tap_count_down - dt
            if self.waiting_tap_count_down <= 0 then
                -- 显示 skip 按钮
                guide_helper:showSkipButtonCallback(true)
            end
        end
    end,

    -- 略过引导
    skipGuide = function(self)

        -- 超时，算引导完成
        log.info("### onGuideOverTime")
        guide_helper:onGuideOverTime()

        if self.coroutine_id then
            app:stopCoroutine(self.coroutine_id)
            self.coroutine_id = nil
        end
        
        if frame_manager:isDisplay("dialog") then
            frame_manager:hide("dialog")
        end

        app:logEvent("skip_guide_"..tostring(self.running_guide_id), tostring(self.running_guide_id))
        self:onGuideComplete(self.running_guide_id, true)
    end,

    -- 保存引导数据
    saveGuideData = function(self)
        network_sender:reqSaveGuide(self.complete_guide_id, self.random_guide_complete)
    end,

    -- 加载引导数据
    loadGuideData = function(self, msg)
        self.complete_guide_id = msg.guide.orderGuide
        local rand_guide = GuideInfoPacker.unpack(msg.guide.randGuide)
        self.random_guide_complete = {}

        if rand_guide.Table then
            for k,v in pairs(rand_guide.Table) do
                self.random_guide_complete[v+10000] = true
               -- log.info("## rand guide:"..v+10000)
            end
        end
    end,
}

if not guide_manager then
    guide_manager = guide_manager_class()
end

return guide_manager