local log = logger.get("Game", "LevelGameObject")

-- 关卡游戏对象
return class {
    ctor = function(self, prefab_path, pos, yaw, scale, callback)
        self.go = nil
        self.prefab_path = prefab_path
        self.pos = pos
        self.yaw = yaw
        self.scale = scale or 1
        self.parent = nil   -- 父结点
        self.head_icon_path = nil
        self:reload(callback)
    end,

    extensions = { "level/level_game_object+states.lua"},

    reload = function(self, callback)
        if self.go then
            self.go:destroy()
            self.go = nil
        end
        self.anim_comps = nil
        self.cs_soldier = nil
        
        log.info("### reload:"..self.prefab_path)

        self.go = level_script.pool:instantiate(self.prefab_path)
        self.go.transform.position = self.pos
        if self.yaw then
            self.go.transform.rotation = Quaternion.AngleAxis(self.yaw, Vector3.up)
        end
        if self.scale then
            self.go.transform.localScale = Vector3.New(self.scale, self.scale, self.scale)
        end

        self.transform = self.go.transform

        if self.rotation_value then
            self.transform.rotation = self.rotation_value
        end

        if level.shadow_projector then
            level.shadow_projector:AddCaster(self.transform)
        end

        local load_callback = function()
            self:setHeadIcon(self.head_icon_path)
            log.info("### finish reload:"..self.prefab_path)

            if callback then
                callback(self)
            end
            self.is_loaded = true
        end

        if self.go.isLoaded then
            app:startCoroutine(
                function()
                    if self.parent then
                        while not self.parent.transform do
                            coroutine.yield(nil)
                        end
                        self.go.transform:SetParent(self.parent.transform)
                        self.go.transform.position = self.parent.transform.position
                        self.go.transform.rotation = self.parent.transform.rotation

                    end
                    coroutine.yield(nil)
                    load_callback()
                end
            )
        else
            self.go.onLoaded = load_callback
        end
    end,

    setRotation = function(self, rot)
        self.rotation_value = rot
        self.transform.rotation = rot
    end,

    -- 移动到目标点
    moveTo = function(self, to, speed, callback)
        app:startCoroutine(
            function()
                if self.is_moving then
                    self.is_moving = false
                    coroutine.yield(nil)
                end

                self.is_moving = true

                local go = self.go;
                local from = go.transform.position
                local start_time = Time.time
                local dist = Vector3.Distance(from, to)
                local total_time = dist / speed
                
                self.move_coroutine = app:startCoroutine(
                        function()
                            while self.is_moving and go.transform do

                                local percent = (Time.time-start_time)/total_time
                                self.pos = Vector3.Lerp(from, to, percent)
                                go.transform.position = self.pos
                                if percent >= 1 then
                                    break
                                end
                                coroutine.yield(nil)
                            end

                            if callback and self.is_moving then
                                callback()
                            end

                        end)                   
                end
        )
   
    end,

    -- 停止移动
    stopMove = function(self)
        self.is_moving = false
    end,

    -- 转向目标
    rotateTo = function(self, target, speed, callback)
        app:startCoroutine(
            function()
                if self.is_rotating then
                    self.is_rotating = false
                    coroutine.yield(nil)
                end
                self.is_rotating = true

                local self_trans = self.go.transform
                self.rotate_coroutine = app:startCoroutine(
                    function()
                        while  self.is_rotating and self_trans do
                            local rt = Quaternion.LookRotation(target.position-self_trans.position)
                            self_trans.rotation = 
                                Quaternion.RotateTowards(
                                    self.go.transform.rotation, 
                                    rt,
                                    Time.deltaTime * speed)
                            if self_trans.rotation == rt then
                                break
                            end
                            coroutine.yield(nil)
                        end

                        if self.is_rotating and callback then
                            callback()
                        end

                    end
                )
            end
        )

    end,

    -- 停止旋转
    stopRotate = function(self)
        self.is_rotating = false
    end,

    -- 播放动画
    playAnimation = function(self, anim_name)
        if not self.anim_comps then
            self.anim_comps = self.go:GetComponentsInChildren(UnityEngine.Animation)
        end

        for _,anim in pairs(self.anim_comps.Table) do
            anim:CrossFade(anim_name)
        end
    end,

    playSoldierAnimation = function(self, anim_name)
        if not self.cs_soldier then
            self.cs_soldier = self.go:GetComponentInChildren(Soldier)
        end

        if self.cs_soldier then
            local soldier_cls = require('RTS.rts_soldier')
            local prefix = soldier_cls.getAnimPostfix(self.cs_soldier)
            self.cs_soldier:PlayAnimation(prefix..anim_name)
        else
            log.error("Failed to find soldier component!")
        end
    end,

    -- 找组件
    GetComponentInChildren = function(self, type, includeInactive)
        return self.go:GetComponentInChildren(type, includeInactive)
    end,

    findChild = function(self, name)
        return level_script:findChild(self.transform, name)
    end,

    -- 设置头顶 icon
    setHeadIcon = function(self, path)
        --log.info("### setHeadIcon:"..tostring(path))
        self.head_icon_path = path
        if not path then
            local child = level_script:findChild(self.transform, "HeadIcon")
            if child then
                child.gameObject:SetActive(false)
            end
        else
            local child = level_script:findChild(self.transform, "HeadIcon")
            if child then
                child.gameObject:SetActive(true)
                local img = child.gameObject:GetComponent(UnityEngine.UI.Image)
                img:SetSprite(path)
            else
                log.error("Failed to find HeadIcon Node:"..self.prefab_path)
            end
        end
    end,

    SetActive = function(self, is_active)
        self.go:SetActive(is_active)
    end,

    update = function(self, dt)
        if self.state_manager then
            self.state_manager:update(dt)
        end
    end,

    destroy = function(self)
        if level then
            if level.shadow_projector then
                level.shadow_projector:RemoveCaster(self.transform)
            end
        end
        
        if self.go then
            self.go:destroy()
            self.go = nil
        end
        self.is_loaded = false
        self.is_moving = false
        self.is_rotating = false
    end,

}