local log = logger.get("Game", "Reconnector")

-- 重连
local reconnector_cls = class{
    states = {
        UNKNOW = 0,    -- 进行连接
        CONNECTING = 1, -- 连接中
        LOGIN = 2,      -- 登陆
        SELECT_ROLE = 3,
        SUCCEED = 4,
        FAILED = 5,
        MAX = 6,
    },

    ctor = function(self)
        self.is_reconnecting = false    -- 是否在重连中
    end,

    -- 执行重连总入口
    reconnect = function(self)
        if self.is_reconnecting then
            -- 正在连接过程中，忽略该请求
            return
        end

        self.is_reconnecting = true
        self.is_failed = false
        self.curr_game_state = game_director.state_name
        self.countdown = 6.0        -- 倒计时

        if not self.orig_time_scale then
            self.orig_time_scale = Time.timeScale
            Time.timeScale = 0.0001
        end

        self.sub_state = self.states.UNKNOW
        self.frame = frame_manager:show('reconnect')
        network:close()
        network:registerMessageHandler(self)
    end,

    update = function(self)
        if not self.is_reconnecting then
            return
        end

        self.countdown = self.countdown - Time.unscaledDeltaTime
        if self.countdown < 0 then
            -- 重连超时失败
            self:failed()
        end

        if self.frame then
            self.frame:setCountdown(math.floor(self.countdown))
        end

        if self.sub_state == self.states.UNKNOW then
            -- 初始状态，待网络关闭就发起连接
            if network:isClosed() then
                self:connect()
            end
        elseif self.sub_state == self.states.CONNECTING then
            -- 连接中,连接成功后进行登陆
            if network:isConnected() then
                self:login()
            end
        end
    end,

    -- 进行连接，网络必须 closed 状态才能调用
    connect = function(self)
        if self.sub_state == self.states.UNKNOW or self.sub_state == self.states.FAILED then
            if not network:isClosed() then
                log.error('connect but network is not close!')
                return false
            end

            log.info("#1 connect!!")
            -- 使用原来的 ip 和 port 连接
            if network:connect(network.ip, network.port) then
                self.sub_state = self.states.CONNECTING -- 连接中
            else
                log.error('Reconnect failed, ip:'..network.ip..', port:'..network.port)
                self:failed()
            end
        end
    end,

    -- 登陆
    login = function(self)
        log.info("#2 login!!")
        network_sender:login(game_info.account, true)
        self.sub_state = self.states.LOGIN
    end,

    -- 选择角色
    selectRole = function(self)
        log.info('#3 select role')

        network_sender:selectRole(game_info.role_list[1].playerid)
        game_info.player_id = game_info.role_list[1].playerid

        self.sub_state = self.states.SELECT_ROLE
    end,

    -- 重连成功
    succeed = function(self)
        log.info("# reconnect successed")
        network:unregisterMessageHandler(self)
        Time.timeScale = self.orig_time_scale or 1.0
        self.is_reconnecting = false

        frame_manager:hide("reconnect")
        self.frame = nil
        self:stateRecover()
    end,

    -- 重连失败
    failed = function(self)
        if not self.is_failed then
            self.is_failed = true

            network:unregisterMessageHandler(self)
            frame_manager:hide("reconnect")
            self.frame = nil
            gui_helper:showMessageBox(
                NativeText.GetLangStr(999001),
                function()
                    self.is_reconnecting = false
                    self:reconnect()
                end
            )
        end

        --self.is_reconnecting = false

    end,

    stateRecover = function(self)
        if game_director:isWorldMapState() then
            network_sender:enterWorldMap(protocol_pb.MAP_WORLDMAP)
        elseif game_director:isHomeState() then
            network_sender:reqChapterAchievementList()
        end
    end,

    onMessage = function(self, cmd, msg)
        if cmd == protocol_pb.LOGIN then
            
            if msg.RespCode == 0 then
                -- 登陆成功 
                game_info.role_list = msg.rolelist
                self:selectRole()
            else
                -- 登陆失败
                log.info('# login failed, RespCode:'..msg.RespCode)
                self:failed()
            end
        elseif cmd == protocol_pb.SELECTROLE then
            if msg.RespCode == 0 then
                local player = require('game.player')
                player.id = msg.playerid
                player.name = msg.rolename
                player.props[protocol_pb.PROP_LEVEL] = msg.level
                player.props[protocol_pb.PROP_CUREXP] = msg.curExp
                player.props[protocol_pb.PROP_GOLD] = msg.gold
                player.props[protocol_pb.PROP_DIAMOND] = msg.diamond
                player.props[protocol_pb.PROR_STEEL] = msg.steel
                player.props[protocol_pb.PROP_WOOD] = msg.wood
                player.props[protocol_pb.PROP_BRICK] = msg.brick
                player.props[protocol_pb.PROP_NUCLEAR] = msg.nuclear

                self:succeed()
            else
                log.info('# select rold failed, RespCode:'..msg.RespCode)
                self:failed()
            end
        end
    end,

}
if not reconnector then
    reconnector = reconnector_cls()
end
return reconnector