local log = logger.get("Game", "World Map Helper")

local outerRadius = 10   -- 六边形外半径
local innerRadius = outerRadius * 0.866025404  -- 六边形内半径

local cls = class{

    ctor = function(self)
        self.empty_monster_circles = {}

        config.worldmap_emptymonsters:foreach(
            function(k, v)
                local circle_cfg = config.worldmap_rings[v.ring]
                if circle_cfg then
                    circle_cfg.emptyMonster = k
                    table.insert(self.empty_monster_circles, circle_cfg)
                end
            end
        )
    end,

    setWorldMap = function(self, world_map)
        self.world_map = world_map
    end,

    getHold = function(self, id)
        local hold = nil
        if self.world_map then
            hold = self.world_map.hold_manager:getHoldFromId(id)
        end

        return hold
    end,

    pointEmptyMonster = function(self, pos)
        local circle = self:pointInCircles(pos, self.empty_monster_circles)
        local empty_monster = nil

        if circle then
            empty_monster = config.worldmap_emptymonsters[circle.emptyMonster]
        end
        
        return empty_monster
    end,

    pointInCircles = function(self, pos, circles)
        for _, circle in pairs(circles) do
            if ((pos.x >= circle.outerbottomleftx and pos.x <= circle.innerbottomleftx) or 
                (pos.x >= circle.inneruprightx and pos.x <= circle.outeruprightx)) and
                (pos.y >= circle.outerbottomlefty and pos.y <= circle.outeruprighty) then
                
                return circle
            end
        end

        return nil
    end,

    getOperateType = function(self, hold)
        local type = nil

        if hold.type == hold_types.player then
            type = self:getPlayerBaseOperate(hold)
        elseif hold.type == hold_types.armyLine then
            type = self:getMarchOperate(hold)
        elseif hold.type == hold_types.titanMine then
            type = self:getTitanMineOperate(hold)
        elseif hold.type == hold_types.guild then
            type = self:getGuildHoldOperate(hold)
        end
        
        return type
    end,

    isOwnMarchLine = function(self, info)
        local own_march = false

        local playerIds = info.players or {}
        for k, v in ipairs(playerIds) do
            if v == local_player.id then
                own_march = true
            end
        end

        return info.playerId == local_player.id or own_march
    end,

    isFriendMarchLine = function(self, info)
        return local_guild:isSelfInGuild() and info.familyId == local_guild.id and not self:isOwnMarchLine(info)
    end,
    
    isEnemyMarchLine = function(self, info)
        if not self:isOwnMarchLine(info) and not self:isFriendMarchLine(info) then
            if info.marchAction == common_pb.EMARCH_ATK or 
                info.marchAction == common_pb.EMARCH_MASS_ATK then
                    if self.world_map then
                        for k, v in ipairs(self.world_map.hold_manager.own_hold) do
                            if info.dstInstid == v.id then
                                return true
                            end
                        end
                    end
            end
        end

        return false
    end,

    getPlayerBaseOperate = function(self, hold)
        if self:isOwnHold(hold) then
            return opreator_type.own_base
        elseif self:isOwnGuild(hold) then
            return opreator_type.friend_base
        else
            return opreator_type.other_base
        end
    end,

    getMarchOperate = function(self, hold)
        local type = nil
        
        if self:isOwnMarchLine(hold.info) then
            if hold.info.action == common_pb.EMARCH_GO_HOME then
                type = opreator_type.march_back
            else
                if hold.info.action == common_pb.EMARCH_MASS_ATK then
                    if local_player.id == hold.info.playerId then
                        type = opreator_type.mass_creator
                    else
                        type = opreator_type.mass_others
                    end
                else
                    type = opreator_type.march_go
                end
            end
        end

        return type
    end,

    getGuildHoldOperate = function(self, hold)
        local type = nil

        if self:isOwnGuild(hold) then
            if hold.info.state == 0 then
                type = opreator_type.own_guild_base_building
            elseif hold.info.state == 1 then
                if self:ownInBuilding(hold) then
                    type = opreator_type.own_guild_base_support
                else
                    type = opreator_type.own_guild_base_building
                end
            elseif hold.info.state == 2 then
                type = opreator_type.own_guild_base_finished
            elseif hold.info.state == 3 then
                type = opreator_type.own_guild_base_demaged
            end
        else
            if hold.info.state == 0 then
                type = opreator_type.other_guild_base_building
            elseif hold.info.state == 1 then
                type = opreator_type.other_guild_base_building
            elseif hold.info.state == 2 then
                type = opreator_type.other_guild_base_finished
            elseif hold.info.state == 3 then
                type = opreator_type.other_guild_base_finished
            end
        end

        return type
    end,

    ownInBuilding = function(self, hold)
        if hold.info.state == 1 then
            for k, v in ipairs(hold.info.teams) do
                if v.playerId == local_player.id then
                    return true
                end
            end
        end

        return false
    end,

    getTitanMineOperate = function(self, hold)
        if local_player.id == hold.info.playerId then
            return opreator_type.own_titan
        elseif string.len(hold.info.playerId) == 0 then
            return opreator_type.origin_titan
        elseif self:isOwnGuild(hold) then
            return opreator_type.friend_titan
        else
            return opreator_type.enemy_titan
        end
    end,

    isOwnGuild = function(self, hold)
        return local_guild:isSelfInGuild() and local_guild:getId() == hold.info.familyId
    end,

    isGuildHold = function(self, type)
        if type == hold_types.guild or 
            type == hold_types.guild_tech or
            type == hold_types.guild_shop or
            type == hold_types.guild_train or
            type == hold_types.guild_sub_base then

                return true
        end

        return false
    end,

    isOwnHold = function(self, hold)
        local info = hold.info
        return info.playerId == local_player.id
    end,

    isNeutralHold = function(self, hold)
        local info = hold.info
        return (not info.playerId or string.len(info.playerId) < 3) and
                    (not info.familyId or string.len(info.familyId) < 3)
    end,

    isEnemyHold = function(self, hold)
        return not self:isOwnGuild(hold) and not self:isNeutralHold(hold) and not self:isOwnHold(hold)
    end,

    isOwnBase = function(self, id)
        if self.world_map then
            return id == self.world_map.my_hold.id
        end

        return false
    end,

    lookAtPosition = function(self, x, z, cb)
        if self.world_map then
            self.world_map:lookAtPosition(x, z, cb)
        end
    end,

    setCameraPos = function(self, position)
        if self.world_map then
            local x, z = self.world_map.cs_inst:GetXZ(position)
            self.world_map:setCameraPos(x, z)
        end
    end,

    getCellPosition = function(self, x, z)
        if self.world_map then
            local pos = self.world_map.cs_inst:GetCellPos(x, z)
            return pos
        end

        return nil
    end,

    getWorldMapBorder = function(self)
        if self.world_map then
            local max_x, max_z = self.world_map:getWorldMapBorder()
            return max_x, max_z
        end

        return 0, 0
    end,

    getMinings = function(self)
        if self.world_map then
            return self.world_map.cells.mine_circles
        end

        return nil
    end,

    posInMining = function(self, x, z)
        if self.world_map then
            return self.world_map.cells:inMineCircles(x, z, 1)
        end

        return false
    end,

    getEmptyName = function(self, x, z)
        local cfg = self:pointEmptyMonster({x = x, y = z})
        local str = NativeText.GetLangStr(280095)

        if cfg then
            str = frame_helper:getLangCfg(cfg.emptymonster_name)
        end

        return str
    end,

    getGuideCamera = function(self)
        local cam = GameObject.Find("HexMap/Hex Map Camera/GuideCamera")
        return cam
    end,

    getHoldById = function(self, id)
        local hold_manager = self.world_map.hold_manager
        local hold = hold_manager:getHoldFromId(id)

        return hold
    end,

    getCameraForward = function(self)
        local cam = Camera.main
        local cam_forward = cam.transform.forward
        cam_forward.y = 0.0
        cam_forward:Normalize()

        return cam_forward
    end,

    getCoordDistance = function(self, pos1, pos2)
        local x1, y1 = pos1.x, pos1.z
        local x2, y2 = pos2.x, pos2.z

        local src_pos = {x = x1 - (y1 - (y1%2))*0.5, y = -x1 - (y1 - (y1%2))*0.5, z = y1}
        local det_pos = {x = x2 - (y2 - (y2%2))*0.5, y = -x2 - (y2 - (y2%2))*0.5, z = y2}

        local iDX = math.abs(src_pos.x - det_pos.x);
        local iDY = math.abs(src_pos.y - det_pos.y);
        local iDZ = math.abs(src_pos.z - det_pos.z);

        local dis = math.max(iDX, iDZ, iDY)

        return dis
    end,

    getFuelExpend = function(self, pos)
        local position = self.world_map.my_hold.transform.position
        local x, z = self.world_map.cs_inst:GetXZ(position)

        local dis = self:getCoordDistance(pos, {x = x, z = z})

        local pre_expend = config.system_config_config_paras[1].base_fuel_cost

        return dis*pre_expend
    end,

    getTransitionCost = function(self, pos)
        local position = self.world_map.my_hold.transform.position
        local x, z = self.world_map.cs_inst:GetXZ(position)

        local dis = self:getCoordDistance(pos, {x = x, z = z})
        
        if dis <= 33 then
            return 40
        end

        local cost = 40 + dis - 33

        return cost
    end,

    getDistToBase = function(self, pos)
        local base = self.world_map.my_hold
        local base_pos = {x = base.coordX, z = base.coordY}

        local dis = self:getCoordDistance(pos, base_pos)

        return dis
    end,

    corpsSubdued = function(self, team1, team2)
        local corps_nums_1 = {}
        local corps_nums_2 = {}

        for k, v in pairs(team1) do
            local cfg = config.soldier_soldiers[v.cfg_id]
            if cfg then
                corps_nums_1[cfg.type] = (corps_nums_1[cfg.type] or 0) + v.count
            end
        end

        for k, v in pairs(team2) do
            local cfg = config.soldier_soldiers[v.cfg_id]
            if cfg then
                corps_nums_2[cfg.type] = (corps_nums_2[cfg.type] or 0) + v.count
            end
        end

        local a_num_1 = corps_nums_1[1] or 0    -- 队伍1 火箭兵数量
        local b_num_1 = corps_nums_1[2] or 0    -- 队伍1 突击兵数量
        local c_num_1 = corps_nums_1[3] or 0    -- 队伍1 狙击手数量
        local d_num_1 = corps_nums_1[4] or 0    -- 队伍1 蛮兵数量

        local a_num_2 = corps_nums_2[1] or 0    -- 队伍2 火箭兵数量
        local b_num_2 = corps_nums_2[2] or 0    -- 队伍2 突击兵数量
        local c_num_2 = corps_nums_2[3] or 0    -- 队伍2 狙击手数量
        local d_num_2 = corps_nums_2[4] or 0    -- 队伍2 蛮兵数量

        local total_1 = a_num_1 + b_num_1 + c_num_1 + d_num_1
        local total_2 = a_num_2 + b_num_2 + c_num_2 + d_num_2

        local w_a_1 = a_num_1 / (total_1 > 0 and total_1 or 1)
        local w_b_1 = b_num_1 / (total_1 > 0 and total_1 or 1)
        local w_c_1 = c_num_1 / (total_1 > 0 and total_1 or 1)
        local w_d_1 = d_num_1 / (total_1 > 0 and total_1 or 1)

        local w_a_2 = a_num_2 / (total_2 > 0 and total_2 or 1)
        local w_b_2 = b_num_2 / (total_2 > 0 and total_2 or 1)
        local w_c_2 = c_num_2 / (total_2 > 0 and total_2 or 1)
        local w_d_2 = d_num_2 / (total_2 > 0 and total_2 or 1)

        local a_b_factor = config.soldier_paras[1].repress12 or 0
        local b_c_factor = config.soldier_paras[1].repress23 or 0
        local c_a_factor = config.soldier_paras[1].repress31 or 0
        
        a_b_factor = a_b_factor / 10000
        b_c_factor = b_c_factor / 10000
        c_a_factor = c_a_factor / 10000

        --log.info("########## w_a_1: "..w_a_1.." w_b_1: "..w_b_1.." w_c_1: "..w_c_1.." w_d_1: "..w_d_1)
        --log.info("########## w_a_2: "..w_a_2.." w_b_2: "..w_b_2.." w_c_2: "..w_c_2.." w_d_2: "..w_d_2)
        --log.info("########## a_b_factor: "..a_b_factor.." b_c_factor: "..b_c_factor.." c_a_factor: "..c_a_factor)

        local subdued_a = w_a_1 * (w_a_2*0 + w_b_2*a_b_factor + w_c_2*c_a_factor*-1 + w_d_2*0)
        local subdued_b = w_b_1 * (w_a_2*a_b_factor*-1 + w_b_2*0 + w_c_2*b_c_factor + w_d_2*0)
        local subdued_c = w_c_1 * (w_a_2*c_a_factor + w_b_2*b_c_factor*-1 + w_c_2*0 + w_d_2*0)

        return subdued_a + subdued_b + subdued_c
    end,

    bestCorps = function(self, team)
        local a_b_factor = config.soldier_paras[1].repress12 or 0
        local b_c_factor = config.soldier_paras[1].repress23 or 0
        local c_a_factor = config.soldier_paras[1].repress31 or 0

        local corps_nums = {}

        for k, v in pairs(team) do
            local cfg = config.soldier_soldiers[v.cfg_id]
            if cfg then
                corps_nums[cfg.type] = (corps_nums[cfg.type] or 0) + v.count
            end
        end

        local a_num = corps_nums[1] or 0    -- 队伍1 火箭兵数量
        local b_num = corps_nums[2] or 0    -- 队伍1 突击兵数量
        local c_num = corps_nums[3] or 0    -- 队伍1 狙击手数量
        --local d_num = corps_nums[4] or 0    -- 队伍1 蛮兵数量

        local total = a_num + b_num + c_num --+ d_num

        local w_a = a_num / (total > 0 and total or 1)
        local w_b = b_num / (total > 0 and total or 1)
        local w_c = c_num / (total > 0 and total or 1)
        --local w_d = d_num / (total > 0 and total or 1)

        if w_b >= 2/3 - w_a and w_b >= 1/3 then
            return 1
        elseif w_a <= 2/3 - w_b and w_a <= 1/3 then
            return 2
        elseif w_b <= 1/3 and w_a >= 1/3 then
            return 3
        else
            return 1
        end
    end,

    soldierPower = function(self, cfg_id, atk_percent, def_percent, all_atk_percent, all_def_percent)
        local cfg = config.soldier_soldiers[cfg_id]
        local para_cfg = config.soldier_paras[1]

        if cfg then
            local base_atk = cfg.ATK
            local percent = atk_percent[cfg.type] * 0.0001
            local all_percent = all_atk_percent * 0.0001
            local factor = para_cfg.atk_para * 0.0001

            local final_atk = base_atk * (percent + all_percent + 1) * factor
            
            local base_def = cfg.DEF
            percent = def_percent[cfg.type] * 0.0001
            all_percent = all_def_percent * 0.0001
            factor = para_cfg.def_para * 0.0001

            local final_def = base_def * (percent + all_percent + 1) * factor

            return final_atk + final_def
        end

        return 0
    end,

    herosAddition = function(self, heros, is_enemy)
        local atk_percent = {}
        local def_percent = {}

        atk_percent[1] = local_player:getGlobalProp(common_pb.SYS_CORP1_ATK_PERCENT)
        atk_percent[2] = local_player:getGlobalProp(common_pb.SYS_CORP2_ATK_PERCENT)
        atk_percent[3] = local_player:getGlobalProp(common_pb.SYS_CORP3_ATK_PERCENT)
        atk_percent[4] = local_player:getGlobalProp(common_pb.SYS_CORP4_ATK_PERCENT)
        local all_atk_percent = local_player:getGlobalProp(common_pb.SYS_ALL_CORP_ATK_PERCENT)

        def_percent[1] = local_player:getGlobalProp(common_pb.SYS_CORP1_DEF_PERCENT)
        def_percent[2] = local_player:getGlobalProp(common_pb.SYS_CORP2_DEF_PERCENT)
        def_percent[3] = local_player:getGlobalProp(common_pb.SYS_CORP3_DEF_PERCENT)
        def_percent[4] = local_player:getGlobalProp(common_pb.SYS_CORP4_DEF_PERCENT)
        local all_def_percent = local_player:getGlobalProp(common_pb.SYS_ALL_CORP_DEF_PERCENT)

        if is_enemy then
            atk_percent[1] = 0
            atk_percent[2] = 0
            atk_percent[3] = 0
            atk_percent[4] = 0
            all_atk_percent = 0
    
            def_percent[1] = 0
            def_percent[2] = 0
            def_percent[3] = 0
            def_percent[4] = 0
            all_def_percent = 0
        end

        for k, v in pairs(heros) do
            local hero = hero_manager:getHero(v)
            local corp_1_atk = hero:getGlobalProp(common_pb.SYS_CORP1_ATK_PERCENT)
            local corp_2_atk = hero:getGlobalProp(common_pb.SYS_CORP2_ATK_PERCENT)
            local corp_3_atk = hero:getGlobalProp(common_pb.SYS_CORP3_ATK_PERCENT)
            local corp_4_atk = hero:getGlobalProp(common_pb.SYS_CORP4_ATK_PERCENT)
            local all_corp_atk = hero:getGlobalProp(common_pb.SYS_ALL_CORP_ATK_PERCENT)

            local corp_1_def = hero:getGlobalProp(common_pb.SYS_CORP1_DEF_PERCENT)
            local corp_2_def = hero:getGlobalProp(common_pb.SYS_CORP2_DEF_PERCENT)
            local corp_3_def = hero:getGlobalProp(common_pb.SYS_CORP3_DEF_PERCENT)
            local corp_4_def = hero:getGlobalProp(common_pb.SYS_CORP4_DEF_PERCENT)
            local all_corp_def = hero:getGlobalProp(common_pb.SYS_ALL_CORP_DEF_PERCENT)

            atk_percent[1] = atk_percent[1] + corp_1_atk
            atk_percent[2] = atk_percent[2] + corp_2_atk
            atk_percent[3] = atk_percent[3] + corp_3_atk
            atk_percent[4] = atk_percent[4] + corp_4_atk
            all_atk_percent = all_atk_percent + all_corp_atk

            def_percent[1] = def_percent[1] + corp_1_def
            def_percent[2] = def_percent[2] + corp_2_def
            def_percent[3] = def_percent[3] + corp_3_def
            def_percent[4] = def_percent[4] + corp_4_def
            all_def_percent = all_def_percent + all_corp_def
        end

        return atk_percent, def_percent, all_atk_percent, all_def_percent
    end,

    calculateTeamPower = function(self, heros, team, is_enemy)
        local atk_percent, def_percent, all_atk_percent, all_def_percent = self:herosAddition(heros, is_enemy)

        local total_num = 0
        local nums = {}
        local corp_soldiers = {}
        for i=1, 4 do
            nums[i] = 0
            corp_soldiers[i] = {}
        end

        for k, v in pairs(team) do
            local cfg = config.soldier_soldiers[v.cfg_id]

            if cfg then
                total_num = total_num + v.count
                nums[cfg.type] = nums[cfg.type] + v.count
                corp_soldiers[cfg.type][v.cfg_id] = v.count
            end
        end

        if total_num == 0 then
            return 0
        end

        local w_power = 0
        local corp_power = {}

        for type, soldiers in pairs(corp_soldiers) do
            corp_power[type] = 0
            for id, count in pairs(soldiers) do
                local power = self:soldierPower(id, atk_percent, def_percent, all_atk_percent, all_def_percent)
                corp_power[type] = corp_power[type] + power * count
            end
            w_power = w_power + corp_power[type] / total_num
        end

        local para_cfg = config.soldier_paras[1]
        local num_para = para_cfg.num_para * 0.0001
        local final_power = w_power * math.pow(total_num, num_para)
        final_power = math.pow(final_power, para_cfg.Showpower_Powerpara*0.0001)*para_cfg.Showpower_Multipara*0.0001

        final_power = final_power > 1 and final_power or 1

        return math.floor(final_power)
    end,

    teamSoldiersAmount = function(self, heros)
        local base_val = local_player:getGlobalProp(common_pb.SYS_BASE_CORP_NUM)
        local percent_val = local_player:getGlobalProp(common_pb.SYS_BASE_CORP_NUM_PERCENT)

        for k, v in pairs(heros) do
            local hero = hero_manager:getHero(v)
            local base = hero:getGlobalProp(common_pb.SYS_BASE_CORP_NUM)
            local percent = hero:getGlobalProp(common_pb.SYS_BASE_CORP_NUM_PERCENT)

            base_val = base_val + base
            percent_val = percent_val + percent
        end

        local final = math.floor(base_val * (percent_val * 0.0001 + 1))

        return final
    end,

    marchTime = function(self, distance, speed)
        distance = distance * innerRadius * 2

        return distance / (speed > 0 and speed or 1)
    end,

    getBaseCurrPos = function(self)
        local x, z = 0, 0
        if self.world_map then
            local position = self.world_map.my_hold.transform.position
            x, z = self.world_map.cs_inst:GetXZ(position)
        end

        return x, z
    end,

    getHoldByPos = function(self, x, y)
        if self.world_map then
            local hold = self.world_map.hold_manager:getHoldFromPos(x, y)

            return hold
        end

        return nil
    end,

    counterJumpToHold = function(self, x, y)
        if self.world_map then
            local cb = function()
                local hold = self:getHoldByPos(x, y)

                if hold then
                    network_sender:queryWorldMap(hold.id, protocol_pb.MAP_WORLDMAP)
                end
            end

            self.world_map:lookAtPosition(x, y, cb)  
        else
            world_map_info.enter_pos = {}
            world_map_info.enter_pos.x = x
            world_map_info.enter_pos.y = y
            game_director:changeState('loading', {next_state = 'world_map', scene_name = 'hex_map', counterJump = true})
        end
    end,

    cube_directions = {
        Vector3(1, -1, 0), Vector3(1, 0, -1), Vector3(0, 1, -1), 
        Vector3(-1, 1, 0), Vector3(-1, 0, 1), Vector3(0, -1, 1),
    },

    getGuildBaseRangeCells = function(self, x, y, radius)
        local list = {}

        if self.world_map then
            local _x = x - (y - y%2)/2
            local _z = y
            local _y = -_x - _z

            table.insert(list, Vector2(x, y))

            for k=1, radius do
                local cube = Vector3(_x-k, _y, _z+k)
                for i=1, 6 do
                    for j=1, k do
                        cube = cube + self.cube_directions[i]

                        local y_1 = -cube.x - cube.y
                        local x_1 = (cube.x - cube.y) * 0.5

                        local pos = Vector2(x_1, y_1)
                        table.insert(list, pos)
                    end
                end
            end
        end

        return list
    end,

    findNeighborGuild = function(self, hold)
        local list = self.world_map.hold_manager:findNeighborGuild(hold)

        return list
    end,

    setRealGuildBaseRange = function(self, hold)
        local all_cells = self:getGuildBaseRangeCells(hold.coordX, hold.coordY, 4)
        local neighbors = self:findNeighborGuild(hold)

        local repeat_cells = {}

        for _, pos in pairs(all_cells) do
            local add = false
            local idx = self.world_map.cells:getIndex(pos.x, pos.y)

            for _, neighbor in pairs(neighbors) do
                if neighbor.range[idx] then
                    add = true

                    if hold.info.state > 1 and neighbor.info.state > 1 then
                        local early = hold.info.finishtime < neighbor.info.finishtime

                        if early then
                            hold:addRange(idx)
                            neighbor:removeRange(idx)

                            repeat_cells[idx] = neighbor.range[idx]
                        end
                    else
                        if (self:isOwnGuild(hold) or hold.info.state > 1) and neighbor.info.state <= 1 then
                            hold:addRange(idx)
                            neighbor:removeRange(idx)

                            repeat_cells[idx] = neighbor.range[idx]
                        end
                    end

                    break
                end
            end

            if not add then
                hold:addRange(idx)
            end
        end

        return repeat_cells
    end,

    resetRealGuildBaseRange = function(self, hold)
        local neighbors = self:findNeighborGuild(hold, true)

        local cnt = 0
        for k, neighbor in pairs(neighbors) do
            local all_cells = self:getGuildBaseRangeCells(neighbor.coordX, neighbor.coordY, 4)

            local list = {}

            for _, pos in pairs(all_cells) do

                local idx = self.world_map.cells:getIndex(pos.x, pos.y)
                if hold.range[idx] then
                    table.insert( list, idx )
                    neighbor:addRange(idx)
                end

                cnt = cnt + 1
            end

            self.world_map:drawExtendRegions(hold, list)
        end
        
        log.info("############################################################ resetRealGuildBaseRange cnt ", cnt)
    end,

    createTempHold = function(self, cfg_id, type)
        if self.world_map then
            self.world_map:createNewHold(cfg_id, type)
        else
            game_director:changeState('loading', {next_state = 'world_map', scene_name = 'hex_map', building_hold = {cfg_id = cfg_id, type = type}})
        end
    end,

    jumpToHold = function(self, hold_id, x, z)
        if self.world_map then
            app:startCoroutine(
                function()
                    self.world_map:lookAtPosition(x, z)
                    coroutine.yield(WaitForSeconds(0.5))
                    network_sender:queryWorldMap(hold_id, protocol_pb.MAP_WORLDMAP)
                end
            )
        else
            self.jump_hold_info = {id = hold_id, x = x, z = z}
            world_map_info.enter_pos = {}
            world_map_info.enter_pos.x = x
            world_map_info.enter_pos.y = z
            game_director:changeState('loading', {next_state = 'world_map', scene_name = 'hex_map'})
        end
    end,

    junpToArmyLine = function(self, army_info)
        if self.world_map then
            local x, z = self:getArmyLineCurPos(army_info)
            app:startCoroutine(
                function()
                    self.world_map:lookAtPosition(x, z)
                    coroutine.yield(WaitForSeconds(0.5))
                    network_sender:queryWorldMap(army_info.instId, protocol_pb.MAP_WORLDMAP)
                end
            )
        else
            self.jump_army_info = army_info
            game_director:changeState('loading', {next_state = 'world_map', scene_name = 'hex_map'})
        end
    end,

    getArmyLineCurPos = function(self, info)
        local percent = info.elapsedTime / info.totalTime
        local src_pos = self.world_map.cs_inst:GetCellPos(info.srcPosX, info.srcPosY)
        local dest_pos = self.world_map.cs_inst:GetCellPos(info.dstPosX, info.dstPosY)

        local dir = dest_pos - src_pos
        local cur_pos = dir*percent + src_pos

        local x, z = self.world_map.cs_inst:GetXZ(cur_pos)

        return x, z
    end,

    excuteJumpTohold = function(self)
        if not self.jump_hold_info then
            return
        end

        app:startCoroutine(
            function()
                coroutine.yield(WaitForSeconds(0.5))
                network_sender:queryWorldMap(self.jump_hold_info.id, protocol_pb.MAP_WORLDMAP)
                self.jump_hold_info = nil
            end
        )
    end,

    excuteJumpToArmyLine = function(self)
        if not self.jump_army_info then
            return
        end

        local x, z = self:getArmyLineCurPos(self.jump_army_info)
        app:startCoroutine(
            function()
                self.world_map:lookAtPosition(x, z)
                coroutine.yield(WaitForSeconds(0.5))
                network_sender:queryWorldMap(self.jump_army_info.instId, protocol_pb.MAP_WORLDMAP)
                self.jump_army_info = nil
            end
        )
    end,

    destroyTempHold = function(self)
        if self.world_map then
            self.world_map:destroyTempHold()
        end
    end,

    drawGuildBaseRegions = function(self, hold)
        if self.world_map then
            self.world_map:removGuildBaseRegions(hold.range)
            self.world_map:drawGuildBaseRegions(hold)
        end
    end,

    positionToXZ = function(self, pos)
        if self.world_map then
            local x, z = self.world_map.cs_inst:GetXZ(pos)

            return x, z
        end
    end,
}

world_map_helper = cls()

return world_map_helper